import os
import tempfile
from textwrap import dedent

from wasmtime import (  # type: ignore
    Config,
    Engine,
    Linker,
    Module,
    Store,
    WasiConfig,
)

from .exceptions import WasmExecError
from .schema import Result


class WasmExecutor:
    """
    Execute Python code strings using the VMWare Wasm Labs WebAssembly
    Python runtime.

    This is done by using wasmtime to run Python code in a seperate
    WASM-based interpreter. The wasmtime code is sandboxed using chroot
    for added security.

    Fuel in wasmtime is a mechanism that allows limiting the number of instructions
    executed during WebAssembly code execution. It helps prevent infinite loops or
    excessive computations by setting a maximum amount of fuel that can be consumed.
    Each instruction executed consumes a certain amount of fuel. Once the consumed
    fuel reaches the specified limit, execution is halted. The consume_fuel option
    and fuel parameter in the wasm_exec function enable this feature. The Result
    object includes the fuel_consumed attribute, indicating how much fuel was consumed
    during execution.

    Credit to: https://til.simonwillison.net/webassembly/python-in-a-wasm-sandbox
    for the reference implementation.

    :param use_fuel: Whether to limit the execution by fuel consumption (optional).
    :type use_fuel: bool
    :param fuel: The maximum amount of fuel allowed for execution (optional).
    :type fuel: int
    :param runtime_path: Path to a Wasm runtime (optional)
    :type runtime_path: str
    """

    def __init__(
        self,
        use_fuel: bool = False,
        fuel: int = 400_000_000,
        runtime_path: str = "",
    ):
        self.engine_cfg = Config()
        self.engine_cfg.consume_fuel = use_fuel
        self.engine_cfg.cache = True

        self.linker = Linker(Engine(self.engine_cfg))
        self.linker.define_wasi()

        if not runtime_path:
            runtime_path = os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "wasm_runtime",
                    "python-3.11.3.wasm",
                )
            )
        self.python_module = Module.from_file(self.linker.engine, runtime_path)

        self.config = WasiConfig()
        self.config.argv = ("python", "-c", "")

        self.use_fuel = use_fuel
        self.fuel = fuel

    def exec(self, code: str, globals: dict = {}, locals: dict = {}):
        """
        Execute arbitrary code in an isolated Wasm-based Python interpreter

        :param code: The WebAssembly code to execute.
        :type code: str
        :param globals: Allowed global scope
        :type global: dict
        :param locals: Allowed local scope
        :type locals: dict
        :return: The result of the code execution.
        :rtype: Result
        :raises WasmExecError: If an error occurs during code execution.
        """
        exec_code = f'exec("""{dedent(code)}""", {globals}, {locals})'
        self.config.argv = ("python", "-c", exec_code)

        with tempfile.TemporaryDirectory() as chroot:
            out_log = os.path.join(chroot, "out.log")
            err_log = os.path.join(chroot, "err.log")
            self.config.stdout_file = out_log
            self.config.stderr_file = err_log

            store = Store(self.linker.engine)

            if self.use_fuel:
                store.add_fuel(self.fuel)
            store.set_wasi(self.config)
            instance = self.linker.instantiate(store, self.python_module)

            start = instance.exports(store)["_start"]
            mem = instance.exports(store)["memory"]

            try:
                start(store)  # type: ignore
            except Exception:
                with open(err_log) as f:
                    raise WasmExecError(f.read())

            with open(out_log) as f:
                result = f.read()

            if not self.use_fuel:
                fuel_consumed = None
            else:
                fuel_consumed = store.fuel_consumed()

            return Result(
                result,  # type: ignore
                mem.size(store),  # type: ignore
                mem.data_len(store),  # type: ignore
                fuel_consumed,
            )  # type: ignore



class Result:
    def __init__(self, result, mem_size, data_len, consumed):
        self.text = result.strip()
        self.mem_size = mem_size
        self.data_len = data_len
        self.fuel_consumed = consumed



import hashlib

import requests  # type: ignore


class WASMInstallError(Exception):
    pass


def download_file(url: str, file_name: str = "") -> str:
    if not file_name:
        file_name = url.split("/")[-1]

    r = requests.get(url, stream=True)

    with open(file_name, "wb") as f:
        for chunk in r.iter_content(chunk_size=1024):
            # writing one chunk at a time to pdf file
            if chunk:
                f.write(chunk)
    return file_name


def gen_checksum(file_path: str) -> str:
    with open(file_path, "rb") as f:
        bytes = f.read()  # read entire file as bytes
        file_hash = hashlib.sha256(bytes).hexdigest()
        return file_hash


def get_wasm():
    wasm_path = download_file(
        "https://github.com/vmware-labs/webassembly-language-runtimes/releases/download/python%2F3.11.3%2B20230428-7d1b259/python-3.11.3.wasm"
    )
    checksum_file = download_file(
        "https://github.com/vmware-labs/webassembly-language-runtimes/releases/download/python%2F3.11.3%2B20230428-7d1b259/python-3.11.3.wasm.sha256sum"
    )
    with open(checksum_file, "r") as f:
        server_checksum = f.read().split(" ")[0]
    generated_checksum = gen_checksum(wasm_path)
    if generated_checksum != server_checksum:
        raise WASMInstallError(
            "The sha256 checksum of the downloaded file"
            "is different from that of the provided checksum"
        )
